__title__ = 'gogame'
__author__ = 'BlackRaven0405'
__license__ = 'MIT'
__copyright__ = 'Copyright 2022-present BlackRaven0405'
__version__ = '0.0.1'


from .board import *
from .territory import *
from .player import *
from .enum import *
